// ArraySet.java, created Wed Dec  7 13:48:14 2005 by salcianu
// Copyright (C) 2005 Alex Salcianu <salcianu@alum.mit.edu>
// Licensed under the Modified BSD Licence; see COPYING for details.
package jpaul.DataStructs;

import java.util.Set;
import java.util.LinkedHashSet;
import java.util.AbstractSet;
import java.util.Collection;
import java.util.Iterator;

/**
 * <code>ArraySet</code> is an immutable, array-backed set.  It
 * consumes minimal memory; very good for small sets.
 * 
 * @author  Alex Salcianu - salcianu@alum.mit.edu
 * @version $Id: ArraySet.java,v 1.2 2005/12/07 23:58:16 salcianu Exp $ */
public class ArraySet<T> extends AbstractSet<T> {
    
    /** Creates an <code>ArraySet</code> containing the elements
        present in the set <code>set</code>.  The created
        <code>ArraySet</code> contains the elements of
        <code>set</code> at the point this constructor is invoked;
        future changes to <code>set</code> are not reflected by the
        <code>ArraySet</code>. */
    public ArraySet(Set<T> set) {
	this(set, true);
    }

    /** Creates an <code>ArraySet</code> containing the distinct
        elements from the colection <code>coll</code>.  The created
        <code>ArraySet</code> contains the elements of
        <code>set</code> at the point this constructor is invoked;
        future changes to <code>set</code> are not reflected by the
        <code>ArraySet</code>. */
    public ArraySet(Collection<T> coll) {
	this(coll, false);
    }

    // the underlying array of elements
    private final T[] elemArray;


    /** Use only if you suspect that someone is invoking the full
        {@link #ArraySet(Collection,boolean) constructor} with an
        incorrect <code>collHasDistinctElements</code> argument.  If
        <code>true</code>, everytime that argument is
        <code>trues</code>, the code checks that the collection really
        contains distinct elements.  */
    static boolean DEBUG_DISTINCT_ELEMS = false;


    /** Powerful and unsafe constructor: creates an
	<code>ArraySet</code> containing the distinct elements present
	in the colection <code>coll</code>.  If the second parameter
	<code>collHasDistinctElements</code> is <code>true</code>,
	then this constructor assumes that <code>coll</code> contains
	only distinct elements and skips the costly step of
	determining the unique elements from
	<code>coll</code>. <strong>This feature is unsafe and should
	be used with maximal care!</strong>.

	<p>The created <code>ArraySet</code> contains the elements of
        <code>coll</code> at the point this constructor is invoked;
        future changes to <code>coll</code> are not reflected by the
        <code>ArraySet</code>. */
    public ArraySet(Collection<T> coll, boolean collHasDistinctElements) {
	if(DEBUG_DISTINCT_ELEMS && collHasDistinctElements) {
	    checkHasDistinctElements(coll);
	}

	Collection<T> collDistinctElements =
	    collHasDistinctElements ? 
	    ((Collection<T>) coll) :
	    ((Collection<T>) new LinkedHashSet<T>(coll));

	// this cast is very ugly, but what more can we do here?
	elemArray = (T[]) collDistinctElements.toArray();
    }


    // Debug code: check that "coll" does not have any duplicates
    private void checkHasDistinctElements(Collection<T> coll) {
	// keep track of the already seen elements
	NonIterableSet<T> seenElems = new NonIterableSet<T>(coll.size());
	for(T elem : coll) {
	    if(!seenElems.add(elem)) {
		throw new IllegalArgumentException("coll has duplicate element:" + elem);
	    }
	}
    }


    // Re-implement the contains method from AbstractSet, for speed reasons

    public boolean contains(Object o) {
	for(T e : elemArray) {
	    if((o == e) || ((o != null) && o.equals(e))) {
		return true;
	    }
	}
	return false;
    }
    

    // The implementations of the following two methods is required in
    // order to implement a real Set on top of an AbstractSet

    public Iterator<T> iterator() {
	return new ArrayIterator<T>(elemArray);
    }

    public int size() {
	return elemArray.length;
    }

}
