// DisjointSet.java, created Wed Feb 23 13:06:57 2000 by cananian
// Copyright (C) 1999 C. Scott Ananian <cananian@alumni.princeton.edu>
// Licensed under the Modified BSD Licence; see COPYING for details.
// Note: Taken from jutil, with the graceful premission of Scott C. Ananian.
package jpaul.DataStructs;

import java.util.AbstractMap;
import java.util.AbstractSet;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/**
 * <code>DisjointSet</code> is an implementation of disjoint-set forests
 * using the path compression and union-by-rank heuristics to achieve
 * O(m * alpha(m, n)) runtime, where 'm' is the total number of
 * operations, 'n' is the total number of elements in the set, and
 * 'alpha' denotes the *extremely* slowly-growing inverse Ackermann
 * function.
 * 
 * @author  C. Scott Ananian <cananian@alumni.princeton.edu>
 * @version $Id: DisjointSet.java,v 1.2 2005/10/31 22:30:12 salcianu Exp $
 */
public class DisjointSet<E>  {
    private final Map<E,Node<E>> elmap = new LinkedHashMap<E,Node<E>>();

    /** Creates a <code>DisjointSet</code>. */
    public DisjointSet() { }

    /** 
	Unites the dynamic sets that contain <code>o1</code> and
	<code>o2</code>, say S1 and S2, into a new set that is the
	union of these two sets.  The two sets are assumed to be
	disjoint prior to the operation.  The representative of the
	resulting set is the representative of either S1 or S2; if
	both S1 and S2 were previously singletons, the representative
	of S1 union S2 is the representative of S2. 
	
	@return The representative of the resulting set.
     */
    public E union(E o1, E o2) { // this is UNION
	Node<E> x = elmap.get(o1);
	if (x==null) x=_make_set(o1);
	Node<E> y = elmap.get(o2);
	if (y==null) y=_make_set(o2);
	assert !x.equals(y) : "Sets assumed to be disjoint";
	return _union(x, y).element;
    }

    /** Returns the representative of the (unique) set containing
     *  <code>o</code>. */
    public E find(E o) { // this is FIND-SET
	Node<E> x = elmap.get(o);
	if (x==null) return o;
	return _find_set(x).element;
    }

    /** Determines if there is a set of more than one element containing
     *  <code>o</code>. */
    public boolean contains(Object o) {
	return elmap.containsKey(o);
    }

    // these are the routines according to CLR
    private Node<E> _make_set(E o) {
	assert !elmap.containsKey(o);
	Node<E> x = new Node<E>(o);
	elmap.put(o, x);
	return x;
    }

    private Node<E> _find_set(Node<E> x) {
	if (x.parent != x)
	    x.parent = _find_set(x.parent);
	return x.parent;
    }

    // union too nodes - also return the representative of the resulting set
    private Node<E> _union(Node<E> x, Node<E> y) {
	return _link(_find_set(x), _find_set(y));
    }

    private Node<E> _link(Node<E> x, Node<E> y) {
	if (x.rank > y.rank) {
	    y.parent = x;
	    return x;
	}
	else {
	    x.parent = y;
	    if (x.rank == y.rank)
		y.rank++;
	    return y;
	}
    }

    /** Returns an unmodifiable <code>Map</code> view of the disjoint
     *  set, where every element is mapped to its canonical representative.
     */
    public Map<E,E> asMap() {
	return new AbstractMap<E,E>() {
	    public boolean containsKey(Object key) {
		return elmap.containsKey(key);
	    }
	    // XXX: returns identity mapping for objects not in set.
	    public E get(Object key) { return find((E)key); }
	    public Set<Map.Entry<E,E>> entrySet() {
		final Set<E> objects = elmap.keySet();
		return new AbstractSet<Map.Entry<E,E>>() {
		    public int size() { return objects.size(); }
		    public Iterator<Map.Entry<E,E>> iterator() {
			final Iterator<E> objit = objects.iterator();
			return new Iterator<Map.Entry<E,E>>() {
			    public boolean hasNext(){ return objit.hasNext(); }
			    public Map.Entry<E,E> next() {
				final E key = objit.next();
				return new AbstractMapEntry<E,E>() {
				    public E getKey() { return key; }
				    public E getValue() {
					// note deferred for efficiency.
					return find(key);
				    }
				};
			    }
			    public void remove() {
				throw new UnsupportedOperationException();
			    }
			};
		    }
		};
	    }
	};
    }

    // re-implemented here, to avoid a dependency on jutil.AbstractMapEntry
    private static abstract class AbstractMapEntry<K,V> implements Map.Entry<K,V> {
	public abstract K getKey();
	public abstract V getValue();
	public V setValue(V newValue) { throw new UnsupportedOperationException(); }
	public int hashCode() {
	    return 
		3 * (getKey()   == null ? 0 : getKey().hashCode()) +
		5 * (getValue() == null ? 0 : getValue().hashCode());
	}
	public boolean equals(Object obj) {
	    if(obj == this) return true;
	    if(obj == null) return false;
	    if(!(obj instanceof Map.Entry/*<K,V>*/))
		return false;
	    Map.Entry<K,V> e2 = (Map.Entry<K,V>) obj;
	    return 
		DSUtil.checkEq(this.getKey(), e2.getKey()) &&
		DSUtil.checkEq(this.getKey(), e2.getKey());
	}
	public String toString() {
	    return "<" + getKey() + "," + getValue() + ">";
	}
    }

    /** Returns a human-readable representation of the DisjointSet. */
    public String toString() {
	Relation<E,E> rel = new MapSetRelation<E,E>();
	for (Node<E> n : elmap.values()) {
	    Node<E> r = _find_set(n);
	    if (n!=r) rel.add(r.element, n.element);
	}
	return rel.toString();
    }

    // node representation.
    private static class Node<E> {
	Node<E> parent;
	final E element;
	int rank;
	Node(E element) {
	    this.parent = this; this.element = element; this.rank = 0;
	}
    }

    /** Self-test method. */
    public static void main(String[] args) {
	DisjointSet<String> ds = new DisjointSet<String>();
	String a="a", b="b", c="c", d="d", e="e", f="f", g="g", h="h";
	assert !ds.contains(a) && !ds.contains(b) && !ds.contains(c);
	assert ds.find(a)==a && ds.find(b)==b && ds.find(c)==c;
	assert !ds.contains(a) && !ds.contains(b) && !ds.contains(c);
	ds.union(e, c); ds.union(b, h); ds.union(h, c);
	assert ds.find(e)==ds.find(c) && ds.find(h)==ds.find(e);
	assert ds.find(b)==ds.find(c) && ds.find(b)!=ds.find(a);
	ds.union(d, f); ds.union(g, d);
	assert ds.find(d)==ds.find(f) && ds.find(f)==ds.find(g);
	assert ds.find(d)!=ds.find(c) && ds.find(d)!=ds.find(a);
	ds.union(c, f);
	assert ds.find(e)==ds.find(f);
	assert ds.find(a)==a;
	System.err.println("PASSED.");
    }
}
