// WorkQueue.java, created Tue Oct  4 18:29:33 2005 by salcianu
// Copyright (C) 2005 Alexandru Salcianu <salcianu@alum.mit.edu>
// Licensed under the Modified BSD Licence; see COPYING for details.
package jpaul.DataStructs;

import java.util.Queue;
import java.util.AbstractQueue;
import java.util.Iterator;
import java.util.ConcurrentModificationException;

/**
 * <code>WorkQueue</code>
 * 
 * @author  Alexandru Salcianu - salcianu@alum.mit.edu
 * @version $Id: WorkQueue.java,v 1.1 2005/10/05 16:36:19 salcianu Exp $
 */
public class WorkQueue<T> extends AbstractQueue<T> {
    
    public WorkQueue() { this(128); }

    /** Creates a <code>WorkQueue</code>. */
    public WorkQueue(int blockCapacity) {
	this.blockCapacity = blockCapacity;

        headNode = new Node<T>(blockCapacity);
	headNode.next = headNode;
	headNode.prev = headNode;
	headIdx = 0;

	tailNode = headNode;
	tailIdx = 0;
    }

    private final int blockCapacity;

    private static class Node<T> {
	Node(int blockCapacity) {
	    content = new Object[blockCapacity];
	}
	Node<T> next;
	Node<T> prev;
	Object[] content;
    };

    private Node<T> headNode;
    private int headIdx;

    private Node<T> tailNode;
    private int tailIdx;

    private int size = 0;
    private int version = 0;

    public boolean offer(T elem) {
	tailNode.content[tailIdx] = elem;
	tailIdx++;
	if(tailIdx == blockCapacity) {
	    if(tailNode.next == headNode) {
		Node<T> newNode = new Node<T>(blockCapacity);
		newNode.next  = headNode;
		headNode.prev = newNode;
		newNode.prev  = tailNode;
		tailNode.next = newNode;
	    }
	    tailNode = tailNode.next;
	    tailIdx = 0;
	}
	size++;
	version++;
	return true;
    }

    public boolean isEmpty() {
	return size == 0;
    }

    public T peek() {
	if(isEmpty()) return null;
	return (T) headNode.content[headIdx];
    }

    public T poll() {
	if(isEmpty()) return null;
	T elem = (T) headNode.content[headIdx];
	// possibly enable some GC
	headNode.content[headIdx] = null;
	headIdx++;
	if(headIdx == blockCapacity) {
	    headIdx = 0;
	    headNode = headNode.next;
	}
	size--;
	version++;
	return elem;
    }


    public int size() {
	return size;
    }


    public Iterator<T> iterator() {
	return new Iterator<T>() {
	    private final int expectedVersion = version;
	    private Node<T> currNode = headNode;
	    private int currIdx = headIdx;
	    public boolean hasNext() {
		return (currNode != tailNode) || (currIdx != tailIdx);
	    }
	    public T next() {
		if(version != expectedVersion) {
		    throw new ConcurrentModificationException();
		}
		T elem = (T) currNode.content[currIdx];
		currIdx++;
		if(currIdx == blockCapacity) {
		    currNode = currNode.next;
		    currIdx = 0;
		}
		return elem;
	    }
	    public void remove() {
		throw new UnsupportedOperationException();
	    }
	};
    }


    public static void main(String[] args) {
	Queue<String> q = new WorkQueue<String>(3);
	q.offer("a");
	q.offer("b");
	System.out.println("q = " + q.toString());
	q.offer("c");
	System.out.println("q = " + q.toString());
	q.offer("d");
	q.offer("e");
	System.out.println("q = " + q.toString());

	assert q.poll().equals("a");
	assert q.poll().equals("b");
	System.out.println("q = " + q.toString());
	assert q.poll().equals("c");
	System.out.println("q = " + q.toString());

	q.offer("f");
	q.offer("g");
	System.out.println("q = " + q.toString());
	q.offer("h");
	System.out.println("q = " + q.toString());
	

	assert q.poll().equals("d");
	System.out.println("q = " + q.toString());
	assert q.poll().equals("e");
	System.out.println("q = " + q.toString());
	q.remove();
    }

}
